//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <vector>
#include "Eigen/Dense"
#include "pinocchio/algorithm/crba.hpp"


using namespace Eigen; 


/**
 * \brief Compute a 3D acceleration from 6D spatial acceleration
 * 
 * \param ddr spatial acc linear part
 * \param w spatial acc rotational part
 * \param dr spatial velocity linear part
**/
Vector3d computeAccFromSpatial(const Vector3d& ddr, const Vector3d& w, const Vector3d& dr);


/**
* \brief Compute the relative contact points from foot center of a Talos foot.
* 
* Order is clockwise, starting from bottom left (looking at a forward pointing foot from above).
* Expressed in local foot coordinates. 
**/
std::vector<Vector3d> contacts_from_footrect_center();


/**
* \brief Compute the wrench at the end effector center expressed in world coordinates.

* Each contact force (at a foot for instance) is represented in MAPI as a set of 
* 4 forces at the corners of the contact polygone (foot -> rectangle). These forces,
* as the other quantities, are expressed in world coordinates. From this 4 3D forces, 
* we can compute the 6D wrench at the center of the origin point of the contacts vectors.

* \param contacts std vector of 3D contact forces
* \param cforces 12D corner forces vector ordered as [fx1, fy1, fz1, fx2, fy2... fz4], same order as contacts
* \param wRl rotation matrix, orientation from world frame to foot frame
**/
Matrix<double, 6, 1> contact_force_to_wrench(const std::vector<Vector3d>& contacts, 
                                             const Matrix<double, 12, 1>& cf12);


/**
* \brief Compute the centroidal angular inertia used to compute the angular momentum.

* \param model: pinocchio robot model used
* \param data: pinocchio robot data used
* \param q_static: configuration of the robot with free flyer pose at origin (local base frame = world frame)
* \param b_p_bc: measure of the CoM position relative to the base
**/
Matrix3d compute_Iw(pinocchio::Model& model, 
                    pinocchio::Data& data, 
                    VectorXd& q_static, 
                    Vector3d& b_p_bc);
